@@ -1,7 +1,7 @@
 lowercase_letters = """abcdefghijklmnopqrstuvwxyz"""
 uppercase_letters = """ABCDEFGHIJKLMNOPQRSTUVWXYZ"""
 numbers = """0123456789"""
-symbols = """!#\$%&'()*+,-./:;"@[]^_`{|}~>?="""
+symbols = """!#\$%&'()*+,-./:;"@[]^_`{|}~>?= """
 symbols_set = set(symbols)
 all_chars = set(lowercase_letters + uppercase_letters + numbers + symbols + " " + "\<")
 other_symbols = {"<space>": " ",
@@ -14,6 +14,12 @@ def get_font():
 def get_heading_font():
     return ("Calibri", 10, "normal")
 
+def is_iterable(o):
+    try:
+        for e in o:
+            return True
+    except:
+        return False
 
 theme_light_blue = {
 'popup_menu_fg': "gray2",
@@ -66,19 +72,19 @@ theme_light_green = {
 'popup_menu_highlight_fg': "black",
 'index_hidden_rows_expander_bg': "gray30",
 'header_hidden_columns_expander_bg': "gray30",
-'header_bg': "#edebe9",
+'header_bg': "#ECECEC",
 'header_border_fg': "#ababab",
 'header_grid_fg': "#ababab",
 'header_fg': "#373737",
 'header_selected_cells_bg': "#d6d4d2",
 'header_selected_cells_fg': "#217346",
-'index_bg': "#edebe9",
+'index_bg': "#ECECEC",
 'index_border_fg': "#ababab",
 'index_grid_fg': "#ababab",
 'index_fg': "black",
 'index_selected_cells_bg': "#d6d4d2",
 'index_selected_cells_fg': "#217346",
-'top_left_bg': "#edebe9",
+'top_left_bg': "#ECECEC",
 'top_left_fg': "#b7b7b7",
 'top_left_fg_highlight': "#5f6368",
 'table_bg': "#fcfcfc",
@@ -169,10 +175,10 @@ theme_black = {
 'top_left_fg_highlight': "#ffb867",
 'table_bg': "#000000",
 'table_grid_fg': "#505054",
-'table_fg': "#c1c1c1",
+'table_fg': "#F2F2F2",
 'table_selected_cells_border_fg': "#bdbdbd",
 'table_selected_cells_bg': "#313034",
-'table_selected_cells_fg': "#dbdbdb",
+'table_selected_cells_fg': "#FFFFFF",
 'resizing_line_fg': "#ffb867",
 'drag_and_drop_bg': "#ecf0f2",
 'outline_color': "gray95",
@@ -182,10 +188,10 @@ theme_black = {
 'index_selected_rows_fg': "#ffb867",
 'table_selected_rows_border_fg': "#bdbdbd",
 'table_selected_rows_bg': "#313034",
-'table_selected_rows_fg': "#dbdbdb",
+'table_selected_rows_fg': "#FFFFFF",
 'table_selected_columns_border_fg': "#bdbdbd",
 'table_selected_columns_bg': "#313034",
-'table_selected_columns_fg': "#dbdbdb"
+'table_selected_columns_fg': "#FFFFFF"
 }
 
 theme_dark_blue = {
